import "dart:math";

class Hex {
  static List directions = [
    new Hex( 1, 0), new Hex( 1,-1), new Hex(0, -1),
    new Hex(-1, 0), new Hex(-1, 1), new Hex(0,  1)
  ];

  final int _q;
  final int _r;

  @override
  toString() => "($q, $r)";

  get q => _q;
  get r => _r;
  get s => -_q - _r;

  int distance(Object other) {
    assert(other is Hex);
    var a = (q - other.q).abs();
    var b = (r - other.r).abs();
    var c = (a - b).abs();
    return max(a, max(b, c));
  }

  Hex neighbor(int direction) {
    assert(0 <= direction && direction < 6);
    return this + directions[direction];
  }

  Point toPixel(Layout layout) {
    Orientation M = layout.orientation;
    return new Point(
      (M.f0 * q + M.f1 * r) * layout.size.x,
      (M.f2 * q + M.f3 * r) * layout.size.y
    );
  }

  Point cornerOffset(Layout layout, int corner) {
    Point size = layout.size;
    double angle = 2.0 * PI * (corner + layout.orientation.startAngle) / 6;
    return new Point(
      layout.origin.x + size.x * cos(angle), 
      layout.origin.y + size.y * sin(angle));
  }

  List<Point> corners(Layout layout) {
    var _corners = new List();
    Point center = toPixel(layout);
    for (int i = 0; i < 6; i++) {
      var offset = cornerOffset(layout, i);
      _corners.add(new Point(center.x + offset.x, center.y + offset.y));
    }
    return _corners;
  }

  bool operator==(Object other) =>
    other is Hex && q == other.q && r == other.r;

  Hex operator+(Object other) {
    assert(other is Hex);
    return new Hex(q + other.q, r + other.r);
  }

  Hex operator-(Object other) {
    assert(other is Hex);
    return new Hex(q - other.q, r - other.r);
  }

  Hex operator*(int k) => new Hex(q * k, r * k);

  Hex(this._q, this._r);

  factory Hex.fromPoint(Layout layout, Point p) {
    var size = layout.size,
        x, z, y, q, r,
        rx, ry, rz,
        x_diff,
        y_diff,
        z_diff;

    if (layout.flat) {
      q = (p.x * 2/3 / size.x);
      r = (-p.x / 3 + sqrt(3)/3 * p.y) / size.y;
    } else {
      q = (p.x * sqrt(3) / 3 - p.y / 3) / size.x;
      r = p.y * 2 / 3 / size.y;
    }

    // convert to cube coordinates
    x = q;
    z = r;
    y = -x - z;
    rx = (x).round();
    ry = (y).round();
    rz = (z).round();
    
    x_diff = (rx - x).abs();
    y_diff = (ry - y).abs();
    z_diff = (rz - z).abs();
    
    if (x_diff > y_diff && x_diff > z_diff) {
      rx = -ry-rz;
    } else if (y_diff > z_diff) {
      ry = -rx-rz;
    } else {
      rz = -rx-ry;
    }

    // convert to axial
    q = rx; 
    r = rz;
    return new Hex(q, r);
  }
}


class Orientation {
  final double f0, f1, f2, f3;
  final double b0, b1, b2, b3;
  final double startAngle;
  Orientation(this.f0, this.f1, this.f2, this.f3,
    this.b0, this.b1, this.b2, this.b3, this.startAngle);
}

class Layout {
  final Orientation orientation;
  final Point size;
  final Point origin;
  final bool flat;
  
  factory Layout(Point size, Point origin) =>
    new Layout.pointy(size, origin);

  Layout.orient_pointy(this.size, this.origin) :
    flat = false,
    orientation = new Orientation(
      sqrt(3.0), sqrt(3.0) / 2.0, 0.0, 3.0 / 2.0,
      sqrt(3.0) / 3.0, -1.0 / 3.0, 0.0, 2.0 / 3.0, 0.5);

  Layout.orient_flat(this.size, this.origin) :
    flat = true,
    orientation = new Orientation(
      3.0 / 2.0, 0.0, sqrt(3.0) / 2.0, sqrt(3.0),
      2.0 / 3.0, 0.0, -1.0 / 3.0, sqrt(3.0) / 3.0, 0.0);
}